from enum import IntEnum


class AudioType(IntEnum):
    UNKNOWN = 0
    ACC = 1
    AIFF = 2
    IT = 10
    MOD = 12
    MPEG = 13
    OGGVORBIS = 14
    S3M = 17
    WAV = 20
    XM = 21
    XMA = 22
    VAG = 23
    AUDIOQUEUE = 24


class AudioCompressionFormat(IntEnum):
    PCM = 0
    Vorbis = 1
    ADPCM = 2
    MP3 = 3
    VAG = 4
    HEVAG = 5
    XMA = 6
    AAC = 7
    GCADPCM = 8
    ATRAC9 = 9


AUDIO_TYPE_EXTEMSION = {
    AudioType.ACC: ".m4a",
    AudioType.AIFF: ".aif",
    AudioType.IT: ".it",
    AudioType.MOD: ".mod",
    AudioType.MPEG: ".mp3",
    AudioType.OGGVORBIS: ".ogg",
    AudioType.S3M: ".s3m",
    AudioType.WAV: ".wav",
    AudioType.XM: ".xm",
    AudioType.XMA: ".wav",
    AudioType.VAG: ".vag",
    AudioType.AUDIOQUEUE: ".fsb",
    AudioCompressionFormat.PCM: ".fsb",
    AudioCompressionFormat.Vorbis: ".fsb",
    AudioCompressionFormat.ADPCM: ".fsb",
    AudioCompressionFormat.MP3: ".fsb",
    AudioCompressionFormat.VAG: ".vag",
    AudioCompressionFormat.HEVAG: ".vag",
    AudioCompressionFormat.XMA: ".wav",
    AudioCompressionFormat.AAC: ".m4a",
    AudioCompressionFormat.GCADPCM: ".fsb",
    AudioCompressionFormat.ATRAC9: ".at9",
}



import os
from typing import Dict, List, Tuple

from .TypeTreeNode import TypeTreeNode

try:
    from TypeTreeGeneratorAPI import TypeTreeGenerator as TypeTreeGeneratorBase  # pyright: ignore[reportAssignmentType]
except ImportError:

    class TypeTreeGeneratorBase:
        def __init__(self, unity_version: str):
            raise ImportError("TypeTreeGeneratorAPI isn't installed!")

        def load_dll(self, dll: bytes): ...
        def load_il2cpp(self, il2cpp: bytes, metadata: bytes): ...
        def get_nodes_as_json(self, assembly: str, fullname: str) -> str: ...
        def get_nodes(self, assembly: str, fullname: str) -> List[TypeTreeNode]: ...


class TypeTreeGenerator(TypeTreeGeneratorBase):
    cache: Dict[Tuple[str, str], TypeTreeNode]

    def __init__(self, unity_version: str, *args, **kwargs):
        super().__init__(unity_version, *args, **kwargs)
        self.cache = {}

    def load_local_game(self, root_dir: str):
        root_files = os.listdir(root_dir)
        data_dir = os.path.join(root_dir, next(f for f in root_files if f.endswith("_Data")))
        if "GameAssembly.dll" in root_files:
            ga_fp = os.path.join(root_dir, "GameAssembly.dll")
            gm_fp = os.path.join(data_dir, "il2cpp_data", "Metadata", "global-metadata.dat")
            ga_raw = open(ga_fp, "rb").read()
            gm_raw = open(gm_fp, "rb").read()
            self.load_il2cpp(ga_raw, gm_raw)
        else:
            self.load_local_dll_folder(os.path.join(data_dir, "Managed"))

    def load_local_dll_folder(self, dll_dir: str):
        for f in os.listdir(dll_dir):
            if not f.endswith(".dll"):
                continue
            fp = os.path.join(dll_dir, f)
            with open(fp, "rb") as f:
                data = f.read()
                self.load_dll(data)

    def get_nodes_up(self, assembly: str, fullname: str) -> TypeTreeNode:
        key = (assembly, fullname)
        if key in self.cache:
            return self.cache[key]

        if not assembly.endswith(".dll"):
            assembly = f"{assembly}.dll"

        base_nodes = self.get_nodes(assembly, fullname)

        node = TypeTreeNode.from_list(
            [
                TypeTreeNode(
                    base_node.m_Level,
                    base_node.m_Type,
                    base_node.m_Name,
                    0,
                    0,
                    m_MetaFlag=base_node.m_MetaFlag,
                )
                for base_node in base_nodes
            ]
        )

        self.cache[key] = node
        return node


__all__ = ("TypeTreeGenerator",)



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from ..classes import MonoBehaviour
from ..classes.ClassIDTypeToClassMap import ClassIDTypeToClassMap
from ..enums import BuildTarget, ClassIDType
from ..exceptions import TypeTreeError
from ..helpers import TypeTreeHelper
from ..helpers.Tpk import get_typetree_node
from ..helpers.TypeTreeNode import TypeTreeNode
from ..helpers.UnityVersion import UnityVersion
from ..streams import EndianBinaryReader, EndianBinaryWriter

if TYPE_CHECKING:
    from ..files.SerializedFile import SerializedFile, SerializedType

T = TypeVar("T")
NodeInput = Union[TypeTreeNode, List[Dict[str, Union[str, int]]]]


class ObjectReader(Generic[T]):
    assets_file: SerializedFile
    reader: EndianBinaryReader
    data: bytes
    version: UnityVersion
    version2: int
    platform: BuildTarget
    path_id: int
    byte_start_offset: Tuple[int, int]
    byte_start: int
    byte_size_offset: Tuple[int, int]
    byte_size: int
    type_id: int
    serialized_type: Optional[SerializedType]
    class_id: int
    type: ClassIDType
    is_destroyed: Optional[int]
    is_stripped: Optional[int]

    # saves where the parser stopped
    # in case that not all data is read
    # and the obj.data is changed, the unknown data can be added again

    def __init__(self, assets_file: SerializedFile, reader: EndianBinaryReader):
        self.assets_file = assets_file
        self.reader = reader
        self.data = b""
        self.version = assets_file.version
        self.version2 = assets_file.header.version
        self.platform = assets_file.target_platform

        header = assets_file.header
        types = assets_file.types

        # AssetStudio ObjectInfo init
        if assets_file.big_id_enabled:
            self.path_id = reader.read_long()
        elif header.version < 14:
            self.path_id = reader.read_int()
        else:
            reader.align_stream()
            self.path_id = reader.read_long()

        if header.version >= 22:
            self.byte_start_offset = (self.reader.real_offset(), 8)
            self.byte_start = reader.read_long()
        else:
            self.byte_start_offset = (self.reader.real_offset(), 4)
            self.byte_start = reader.read_u_int()

        self.byte_start += header.data_offset
        self.byte_header_offset = header.data_offset
        self.byte_base_offset = self.reader.BaseOffset

        self.byte_size_offset = (self.reader.real_offset(), 4)
        self.byte_size = reader.read_u_int()

        self.type_id = reader.read_int()

        if header.version < 16:
            self.class_id = reader.read_u_short()
            self.serialized_type = None
            for typ in types:
                if typ.class_id == self.type_id:
                    self.serialized_type = typ
                    break
        else:
            typ = types[self.type_id]
            self.serialized_type = typ
            self.class_id = typ.class_id

        self.type = ClassIDType(self.class_id)

        if header.version < 11:
            self.is_destroyed = reader.read_u_short()

        if 11 <= header.version < 17:
            script_type_index = reader.read_short()
            if self.serialized_type:
                self.serialized_type.script_type_index = script_type_index

        if header.version == 15 or header.version == 16:
            self.stripped = reader.read_byte()

    def write(self, header, writer: EndianBinaryWriter, data_writer: EndianBinaryWriter):
        if self.assets_file.big_id_enabled:
            writer.write_long(self.path_id)
        elif header.version < 14:
            writer.write_int(self.path_id)
        else:
            writer.align_stream()
            writer.write_long(self.path_id)

        if self.data:
            data = self.data
            # in some cases the parser doesn't read all of the object data
            # games might still require the missing data
            # so following code appends the missing data back to edited objects
            # -> following solution has let to some problems, so it will be removed for now
            # if self.type != ClassIDType.MonoBehaviour:
            #     end_pos = self.byte_start + self.byte_size
            #     if self._read_until and self._read_until != end_pos:
            #         self.reader.Position = self._read_until
            #         data += self.reader.read_bytes(end_pos - self._read_until)
        else:
            self.reset()
            data = self.reader.read(self.byte_size)

        if header.version >= 22:
            writer.write_long(data_writer.Position)
        else:
            writer.write_u_int(data_writer.Position)

        writer.write_u_int(len(data))
        data_writer.write(data)

        writer.write_int(self.type_id)

        if header.version < 16:
            writer.write_u_short(self.class_id)

        if header.version < 11:
            assert self.is_destroyed is not None
            writer.write_u_short(self.is_destroyed)

        if 11 <= header.version < 17:
            assert self.serialized_type is not None
            writer.write_short(self.serialized_type.script_type_index)

        if header.version == 15 or header.version == 16:
            writer.write_byte(self.stripped)

    def set_raw_data(self, data):
        self.data = data
        if self.assets_file:
            self.assets_file.mark_changed()

    def get_class(self) -> Union[Type[T], None]:
        return ClassIDTypeToClassMap.get(self.type)  # type: ignore

    def peek_name(self) -> Union[str, None]:
        """Peeks the name of the object without reading/parsing the whole object."""
        node = self._get_typetree_node()
        peek_node = node.get_name_peek_node()
        if peek_node:
            node, key = peek_node
            return self.parse_as_dict(node, check_read=False)[key]
        else:
            return None

    @property
    def container(self):
        return self.assets_file._container.path_dict.get(self.path_id)

    @property
    def Position(self):
        return self.reader.Position

    @Position.setter
    def Position(self, pos):
        self.reader.Position = pos

    def reset(self):
        self.reader.Position = self.byte_start

    def read(self, check_read: bool = True) -> T:
        obj = self.read_typetree(wrap=True, check_read=check_read)
        self._read_until = self.reader.Position
        return obj  # type: ignore

    def get(self, key, default=None):
        return getattr(self, key, default)

    def __getattr__(self, item: str):
        if hasattr(self.reader, item):
            return getattr(self.reader, item)

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.type.name)

    ###################################################
    #
    #           Typetree Stuff
    #
    ###################################################

    def dump_typetree_structure(
        self,
        nodes: Optional[NodeInput] = None,
        indent: str = "  ",
    ) -> str:
        node = self._get_typetree_node(nodes)
        return node.dump_structure(indent=indent)

    def read_typetree(
        self,
        nodes: Optional[NodeInput] = None,
        wrap: bool = False,
        check_read: bool = True,
    ) -> Union[dict, T]:
        node = self._get_typetree_node(nodes)
        self.reset()
        ret = TypeTreeHelper.read_typetree(
            node,
            self.reader,
            as_dict=not wrap,
            assetsfile=self.assets_file,
            byte_size=self.byte_size,
            check_read=check_read,
        )
        if wrap:
            ret.set_object_reader(self)  # type: ignore
        return ret  # type: ignore

    def save_typetree(
        self,
        tree: Union[dict, T],
        nodes: Optional[NodeInput] = None,
        writer: Optional[EndianBinaryWriter] = None,
    ):
        node = self._get_typetree_node(nodes)
        if not writer:
            writer = EndianBinaryWriter(endian=self.reader.endian)
        TypeTreeHelper.write_typetree(tree, node, writer, self.assets_file)
        data = writer.bytes
        self.set_raw_data(data)
        return data

    def get_raw_data(self) -> bytes:
        pos = self.Position
        self.reset()
        ret = self.reader.read_bytes(self.byte_size)
        self.Position = pos
        return ret

    def _get_typetree_node(
        self,
        node: Optional[NodeInput] = None,
    ) -> TypeTreeNode:
        if isinstance(node, TypeTreeNode):
            return node
        elif isinstance(node, list):
            return TypeTreeNode.from_list(node)
        elif node is not None:
            raise ValueError("nodes must be a list[dict] or TypeTreeNode")

        if self.serialized_type:
            node = self.serialized_type.node
        if not node:
            node = get_typetree_node(self.class_id, self.version)
            if node.m_Type == "MonoBehaviour":
                try:
                    node = self.generate_monobehaviour_node(node)
                except ValueError:
                    pass
        if not node:
            raise TypeTreeError("There are no TypeTree nodes for this object.")
        return node

    # UnityPy 2 syntax early implementation
    def parse_as_object(self, node: Optional[NodeInput] = None, check_read: bool = True) -> T:
        return self.read_typetree(nodes=node, wrap=True, check_read=check_read)  # type: ignore

    def parse_as_dict(self, node: Optional[NodeInput] = None, check_read: bool = True) -> dict[str, Any]:
        return self.read_typetree(nodes=node, wrap=False, check_read=check_read)  # type: ignore

    def patch(
        self,
        obj: Union[dict, T],
        nodes: Optional[NodeInput] = None,
        writer: Optional[EndianBinaryWriter] = None,
    ):
        return self.save_typetree(obj, nodes=nodes, writer=writer)

    # MonoBehaviour specific methods
    def parse_monobehaviour_head(self, mb_node: Optional[TypeTreeNode] = None) -> MonoBehaviour:
        if mb_node is None:
            mb_node = get_typetree_node(ClassIDType.MonoBehaviour, self.version)

        mb = self.read_typetree(nodes=mb_node, wrap=True, check_read=False)
        return cast(MonoBehaviour, mb)

    def generate_monobehaviour_node(self, mb_node: Optional[TypeTreeNode] = None) -> TypeTreeNode:
        env = self.assets_file.environment
        generator = env.typetree_generator
        if generator is None:
            raise ValueError("MonoBehaviour detected, but no typetree_generator set to the environment!")

        monobehaviour = self.parse_monobehaviour_head(mb_node)
        script = monobehaviour.m_Script.deref_parse_as_object()

        if script.m_Namespace != "":
            fullname = f"{script.m_Namespace}.{script.m_ClassName}"
        else:
            fullname = script.m_ClassName

        node = generator.get_nodes_up(script.m_AssemblyName, fullname)
        if node:
            return node
        else:
            raise ValueError(f"Failed to generate MonoBehaviour node for {fullname} of {script.m_AssemblyName}!")



from UnityPy.classes import PPtr
from UnityPy.classes.generated import Component
from UnityPy.classes.legacy_patch import GameObject

class Renderer(Component):
    m_GameObject: PPtr[GameObject]

    def export(self, export_dir: str) -> None: ...



from .AudioClip import AudioClip
from .GameObject import GameObject
from .Mesh import Mesh
from .Renderer import Renderer
from .Shader import Shader
from .Sprite import Sprite
from .Texture2D import Texture2D
from .Texture2DArray import Texture2DArray

__all__ = [
    "AudioClip",
    "GameObject",
    "Mesh",
    "Renderer",
    "Shader",
    "Sprite",
    "Texture2D",
    "Texture2DArray",
]



import json
import os
from io import BytesIO
from pathlib import Path
from typing import Callable, Dict, Iterable, List, Optional, Sequence, Tuple, Union

import UnityPy
from UnityPy.classes import (
    AudioClip,
    Font,
    GameObject,
    Mesh,
    MonoBehaviour,
    Object,
    PPtr,
    Shader,
    Sprite,
    TextAsset,
    Texture2D,
)
from UnityPy.enums.ClassIDType import ClassIDType
from UnityPy.files import SerializedFile


def export_obj(
    obj: Union[Object, PPtr],
    fp: Path,
    append_name: bool = False,
    append_path_id: bool = False,
    export_unknown_as_typetree: bool = False,
    asset_filter: Optional[Callable[[Object], bool]] = None,
) -> List[Tuple[SerializedFile, int]]:
    """Exports the given object to the given filepath.

    Args:
        obj (Object, PPtr): A valid Unity object or a reference to one.
        fp (Path): A valid filepath where the object should be exported to.
        append_name (bool, optional): Decides if the obj name will be appended to the filepath. Defaults to False.
        append_path_id (bool, optional): Decides if the obj path id will be appended to the filepath. Defaults to False.
        export_unknown_as_typetree (bool, optional): If set, then unimplemented objects will be exported
            via their typetree or dumped as bin. Defaults to False.
        asset_filter (func(Object)->bool, optional): Determines whether to export an object. Defaults to all objects.

    Returns:
        list: a list of exported object path_ids
    """
    # figure out export function
    export_func = EXPORT_TYPES.get(obj.type)
    if not export_func:
        if export_unknown_as_typetree:
            export_func = exportMonoBehaviour
        else:
            return []

    # set filepath
    obj = obj.read()

    # a filter that returned True during an earlier extract_assets check can return False now with more info from read()
    if asset_filter and not asset_filter(obj):
        return []

    if append_name:
        fp = os.path.join(
            fp,
            obj.m_Name if getattr(obj, "m_Name") else obj.object_reader.type.name,  # noqa: B009
        )

    fp, extension = os.path.splitext(fp)

    if append_path_id:
        fp = f"{fp}_{obj.object_reader.path_id}"

    # export
    return export_func(obj, fp, extension)


def extract_assets(
    src: Union[Path, BytesIO, bytes, bytearray],
    dst: Path,
    use_container: bool = True,
    ignore_first_container_dirs: int = 0,
    append_path_id: bool = False,
    export_unknown_as_typetree: bool = False,
    asset_filter: Optional[Callable[[Object], bool]] = None,
) -> List[Tuple[SerializedFile, int]]:
    """Extracts some or all assets from the given source.

    Args:
        src (Union[Path, BytesIO, bytes, bytearray]): [description]
        dst (Path): [description]
        use_container (bool, optional): [description]. Defaults to True.
        ignore_first_container_dirs (int, optional): [description]. Defaults to 0.
        append_path_id (bool, optional): [description]. Defaults to False.
        export_unknown_as_typetree (bool, optional): [description]. Defaults to False.
        asset_filter (func(object)->bool, optional): Determines whether to export an object.
            Defaults to all objects.

    Returns:
        List[Tuple[SerializedFile, int]]: [description]
    """
    # load source
    env = UnityPy.load(src)
    exported = []

    export_types_keys = list(EXPORT_TYPES.keys())

    def defaulted_export_index(type: ClassIDType):
        try:
            return export_types_keys.index(type)
        except (IndexError, ValueError):
            return 999

    if use_container:
        container = sorted(env.container.items(), key=lambda x: defaulted_export_index(x[1].type))
        for obj_path, obj in container:
            # The filter here can only access metadata.
            # The same filter may produce a different result later in extract_obj after obj.read()
            if asset_filter is not None and not asset_filter(obj):
                continue
            # the check of the various sub directories is required to avoid // in the path
            obj_dest = os.path.join(
                dst,
                *(x for x in obj_path.split("/")[ignore_first_container_dirs:] if x),
            )
            os.makedirs(os.path.dirname(obj_dest), exist_ok=True)
            exported.extend(
                export_obj(
                    obj,
                    obj_dest,
                    append_path_id=append_path_id,
                    export_unknown_as_typetree=export_unknown_as_typetree,
                    asset_filter=asset_filter,
                )
            )

    else:
        objects = sorted(env.objects, key=lambda x: defaulted_export_index(x.type))
        for obj in objects:
            if asset_filter is not None and not asset_filter(obj):
                continue
            if (obj.assets_file, obj.path_id) not in exported:
                exported.extend(
                    export_obj(
                        obj,
                        dst,
                        append_name=True,
                        append_path_id=append_path_id,
                        export_unknown_as_typetree=export_unknown_as_typetree,
                        asset_filter=asset_filter,
                    )
                )

    return exported


###############################################################################
#                      EXPORT FUNCTIONS                                       #
###############################################################################


def exportTextAsset(obj: TextAsset, fp: str, extension: str = ".txt") -> List[Tuple[SerializedFile, int]]:
    if not extension:
        extension = ".txt"
    with open(f"{fp}{extension}", "wb") as f:
        f.write(obj.m_Script.encode("utf-8", "surrogateescape"))
    return [(obj.assets_file, obj.object_reader.path_id)]


def exportFont(obj: Font, fp: str, extension: str = "") -> List[Tuple[SerializedFile, int]]:
    # TODO - export glyphs
    if obj.m_FontData:
        extension = ".ttf"
        if obj.m_FontData[0:4] == b"OTTO":
            extension = ".otf"
        with open(f"{fp}{extension}", "wb") as f:
            f.write(bytes(obj.m_FontData))
    return [(obj.assets_file, obj.object_reader.path_id)]


def exportMesh(obj: Mesh, fp: str, extension=".obj") -> List[Tuple[SerializedFile, int]]:
    if not extension:
        extension = ".obj"
    with open(f"{fp}{extension}", "wt", encoding="utf8", newline="") as f:
        f.write(obj.export())
    return [(obj.assets_file, obj.object_reader.path_id)]


def exportShader(obj: Shader, fp: str, extension=".txt") -> List[Tuple[SerializedFile, int]]:
    if not extension:
        extension = ".txt"
    with open(f"{fp}{extension}", "wt", encoding="utf8", newline="") as f:
        f.write(obj.export())
    return [(obj.assets_file, obj.object_reader.path_id)]


def exportMonoBehaviour(
    obj: Union[MonoBehaviour, Object], fp: str, extension: str = ""
) -> List[Tuple[SerializedFile, int]]:
    export = None

    if obj.object_reader.serialized_type.node:
        # a typetree is available from the SerializedFile for this object
        export = obj.object_reader.read_typetree()
    elif isinstance(obj, MonoBehaviour):
        # try to get the typetree from the MonoBehavior script
        script_ptr = obj.m_Script
        if script_ptr:
            # looks like we have a script
            script = script_ptr.read()
            # check if there is a locally stored typetree for it
            nodes = MONOBEHAVIOUR_TYPETREES.get(script.m_AssemblyName, {}).get(script.m_ClassName, None)
            if nodes:
                export = obj.object_reader.read_typetree(nodes)
    else:
        export = obj.object_reader.read_typetree()

    if not export:
        extension = ".bin"
        export = obj.object_reader.raw_data
    else:
        extension = ".json"
        export = json.dumps(export, indent=4, ensure_ascii=False).encode("utf8", errors="surrogateescape")
    with open(f"{fp}{extension}", "wb") as f:
        f.write(export)
    return [(obj.assets_file, obj.object_reader.path_id)]


def exportAudioClip(obj: AudioClip, fp: str, extension: str = "") -> List[Tuple[SerializedFile, int]]:
    samples = obj.samples
    if len(samples) == 0:
        pass
    elif len(samples) == 1:
        with open(f"{fp}.wav", "wb") as f:
            f.write(list(samples.values())[0])
    else:
        os.makedirs(fp, exist_ok=True)
        for name, clip_data in samples.items():
            with open(os.path.join(fp, f"{name}.wav"), "wb") as f:
                f.write(clip_data)
    return [(obj.assets_file, obj.object_reader.path_id)]


def exportSprite(obj: Sprite, fp: str, extension: str = ".png") -> List[Tuple[SerializedFile, int]]:
    if not extension:
        extension = ".png"
    obj.image.save(f"{fp}{extension}")
    exported = [
        (obj.assets_file, obj.object_reader.path_id),
        (obj.m_RD.texture.assetsfile, obj.m_RD.texture.path_id),
    ]
    alpha_assets_file = getattr(obj.m_RD.alphaTexture, "assets_file", None)
    alpha_path_id = getattr(obj.m_RD.alphaTexture, "path_id", None)
    if alpha_path_id and alpha_assets_file:
        exported.append((alpha_assets_file, alpha_path_id))
    return exported


def exportTexture2D(obj: Texture2D, fp: str, extension: str = ".png") -> List[Tuple[SerializedFile, int]]:
    if not extension:
        extension = ".png"
    if obj.m_Width:
        # textures can be empty
        obj.image.save(f"{fp}{extension}")
    return [(obj.assets_file, obj.object_reader.path_id)]


def exportGameObject(obj: GameObject, fp: str, extension: str = "") -> List[Tuple[SerializedFile, int]]:
    exported = [(obj.assets_file, obj.object_reader.path_id)]
    refs = crawl_obj(obj)
    if refs:
        os.makedirs(fp, exist_ok=True)
    for ref_id, ref in refs.items():
        # Don't export already exported objects a second time
        # and prevent circular calls by excluding other GameObjects.
        # The other GameObjects were already exported in the this call.
        if (ref.assets_file, ref_id) in exported or ref.type == ClassIDType.GameObject:
            continue
        try:
            exported.extend(export_obj(ref, fp, True, True))
        except Exception as e:
            print(f"Failed to export {ref_id}")
            print(e)
    return exported


EXPORT_TYPES = {
    # following types can include other objects
    ClassIDType.GameObject: exportGameObject,
    ClassIDType.Sprite: exportSprite,
    # following types don't include other objects
    ClassIDType.AudioClip: exportAudioClip,
    ClassIDType.Font: exportFont,
    ClassIDType.Mesh: exportMesh,
    ClassIDType.MonoBehaviour: exportMonoBehaviour,
    ClassIDType.Shader: exportShader,
    ClassIDType.TextAsset: exportTextAsset,
    ClassIDType.Texture2D: exportTexture2D,
}

ASSEMBLY_NAME_DLL = str
CLASS_NAME = str

MONOBEHAVIOUR_TYPETREES: Dict[ASSEMBLY_NAME_DLL, Dict[CLASS_NAME, List[Dict]]] = {}


def crawl_obj(obj: Object, ret: Optional[dict] = None) -> Dict[int, Union[Object, PPtr]]:
    """Crawls through the data struture of the object
    and returns a list of all the components.
    """
    if not ret:
        ret = {}

    if isinstance(obj, PPtr):
        if obj.path_id == 0 and obj.file_id == 0 and obj.index == -2:
            return ret
        try:
            obj = obj.read()
        except AttributeError:
            return ret
    else:
        return ret
    ret[obj.path_id] = obj

    # MonoBehaviour really on their typetree
    # while Object denotes that the class of the object isn't implemented yet
    if isinstance(obj, (MonoBehaviour, Object)):
        data = obj.read_typetree().__dict__.values()
    else:
        data = obj.__dict__.values()

    for value in flatten(data):
        if isinstance(value, (Object, PPtr)):
            if value.path_id in ret:
                continue
            crawl_obj(value, ret)

    return ret


def flatten(seq: Sequence) -> Iterable:
    for elem in list(seq):
        if isinstance(elem, (list, tuple)):
            yield from flatten(elem)
        elif isinstance(elem, dict):
            yield from flatten(elem.values())  # type: ignore
        else:
            yield elem



from . import ArchiveStorageManager, CompressionHelper, ContainerHelper, ImportHelper, TypeTreeHelper, UnityVersion

__all__ = [
    "ArchiveStorageManager",
    "CompressionHelper",
    "ImportHelper",
    "TypeTreeHelper",
    "ContainerHelper",
    "UnityVersion",
]



from enum import IntFlag


class CompressionFlags(IntFlag):
    NONE = 0
    LZMA = 1
    LZ4 = 2
    LZ4HC = 3
    LZHAM = 4


class ArchiveFlagsOld(IntFlag):
    CompressionTypeMask = 0x3F
    BlocksAndDirectoryInfoCombined = 0x40
    BlocksInfoAtTheEnd = 0x80
    OldWebPluginCompatibility = 0x100
    UsesAssetBundleEncryption = 0x200


class ArchiveFlags(IntFlag):
    CompressionTypeMask = 0x3F
    BlocksAndDirectoryInfoCombined = 0x40
    BlocksInfoAtTheEnd = 0x80
    OldWebPluginCompatibility = 0x100
    BlockInfoNeedPaddingAtStart = 0x200
    UsesAssetBundleEncryption = 0x1400  # old: 0x400, new: 0x1000



# based on https://github.com/nesrak1/AssetsTools.NET/blob/dev/AssetsTools.NET.Texture/Swizzle/SwitchSwizzle.cs
from typing import Dict, List, Optional, Tuple, Union

from ..enums import BuildTarget, TextureFormat

GOB_X_TEXEL_COUNT = 4
GOB_Y_TEXEL_COUNT = 8
TEXEL_BYTE_SIZE = 16
TEXELS_IN_GOB = GOB_X_TEXEL_COUNT * GOB_Y_TEXEL_COUNT
GOB_MAP = [(((v >> 3) & 0b10) | ((v >> 1) & 0b1), ((v >> 1) & 0b110) | (v & 0b1)) for v in range(TEXELS_IN_GOB)]


def ceil_divide(a: int, b: int) -> int:
    return (a + b - 1) // b


def deswizzle(
    data: Union[bytes, bytearray, memoryview],
    width: int,
    height: int,
    block_width: int,
    block_height: int,
    texels_per_block: int,
) -> bytearray:
    block_count_x = ceil_divide(width, block_width)
    block_count_y = ceil_divide(height, block_height)
    gob_count_x = block_count_x // GOB_X_TEXEL_COUNT
    gob_count_y = block_count_y // GOB_Y_TEXEL_COUNT
    new_data = bytearray(len(data))
    data_view = memoryview(data)

    for i in range(gob_count_y // texels_per_block):
        for j in range(gob_count_x):
            base_gob_dst_x = j * 4
            for k in range(texels_per_block):
                base_gob_dst_y = (i * texels_per_block + k) * GOB_Y_TEXEL_COUNT
                for gob_x, gob_y in GOB_MAP:
                    dst_offset = ((base_gob_dst_y + gob_y) * block_count_x + (base_gob_dst_x + gob_x)) * TEXEL_BYTE_SIZE
                    new_data[dst_offset : dst_offset + TEXEL_BYTE_SIZE] = data_view[:TEXEL_BYTE_SIZE]
                    data_view = data_view[TEXEL_BYTE_SIZE:]
    return new_data


def swizzle(
    data: Union[bytes, bytearray, memoryview],
    width: int,
    height: int,
    block_width: int,
    block_height: int,
    texels_per_block: int,
) -> bytearray:
    block_count_x = ceil_divide(width, block_width)
    block_count_y = ceil_divide(height, block_height)
    gob_count_x = block_count_x // GOB_X_TEXEL_COUNT
    gob_count_y = block_count_y // GOB_Y_TEXEL_COUNT
    new_data = bytearray(len(data))
    data_view = memoryview(new_data)

    for i in range(gob_count_y // texels_per_block):
        for j in range(gob_count_x):
            base_gob_dst_x = j * 4
            for k in range(texels_per_block):
                base_gob_dst_y = (i * texels_per_block + k) * GOB_Y_TEXEL_COUNT
                for gob_x, gob_y in GOB_MAP:
                    src_offset = ((base_gob_dst_y + gob_y) * block_count_x + (base_gob_dst_x + gob_x)) * TEXEL_BYTE_SIZE
                    data_view[:TEXEL_BYTE_SIZE] = data[src_offset : src_offset + TEXEL_BYTE_SIZE]
                    data_view = data_view[TEXEL_BYTE_SIZE:]

    return new_data


# this should be the amount of pixels that can fit 16 bytes
TEXTURE_FORMAT_BLOCK_SIZE_MAP: Dict[TextureFormat, Tuple[int, int]] = {
    TextureFormat.Alpha8: (16, 1),  # 1 byte per pixel
    TextureFormat.ARGB4444: (8, 1),  # 2 bytes per pixel
    TextureFormat.RGBA32: (4, 1),  # 4 bytes per pixel
    TextureFormat.ARGB32: (4, 1),  # 4 bytes per pixel
    TextureFormat.ARGBFloat: (1, 1),  # 16 bytes per pixel (?)
    TextureFormat.RGB565: (8, 1),  # 2 bytes per pixel
    TextureFormat.R16: (8, 1),  # 2 bytes per pixel
    TextureFormat.DXT1: (8, 4),  # 8 bytes per 4x4=16 pixels
    TextureFormat.DXT5: (4, 4),  # 16 bytes per 4x4=16 pixels
    TextureFormat.RGBA4444: (8, 1),  # 2 bytes per pixel
    TextureFormat.BGRA32: (4, 1),  # 4 bytes per pixel
    TextureFormat.BC6H: (4, 4),  # 16 bytes per 4x4=16 pixels
    TextureFormat.BC7: (4, 4),  # 16 bytes per 4x4=16 pixels
    TextureFormat.BC4: (8, 4),  # 8 bytes per 4x4=16 pixels
    TextureFormat.BC5: (4, 4),  # 16 bytes per 4x4=16 pixels
    TextureFormat.ASTC_RGB_4x4: (4, 4),  # 16 bytes per 4x4=16 pixels
    TextureFormat.ASTC_RGB_5x5: (5, 5),  # 16 bytes per 5x5=25 pixels
    TextureFormat.ASTC_RGB_6x6: (6, 6),  # 16 bytes per 6x6=36 pixels
    TextureFormat.ASTC_RGB_8x8: (8, 8),  # 16 bytes per 8x8=64 pixels
    TextureFormat.ASTC_RGB_10x10: (10, 10),  # 16 bytes per 10x10=100 pixels
    TextureFormat.ASTC_RGB_12x12: (12, 12),  # 16 bytes per 12x12=144 pixels
    TextureFormat.ASTC_RGBA_4x4: (4, 4),  # 16 bytes per 4x4=16 pixels
    TextureFormat.ASTC_RGBA_5x5: (5, 5),  # 16 bytes per 5x5=25 pixels
    TextureFormat.ASTC_RGBA_6x6: (6, 6),  # 16 bytes per 6x6=36 pixels
    TextureFormat.ASTC_RGBA_8x8: (8, 8),  # 16 bytes per 8x8=64 pixels
    TextureFormat.ASTC_RGBA_10x10: (10, 10),  # 16 bytes per 10x10=100 pixels
    TextureFormat.ASTC_RGBA_12x12: (12, 12),  # 16 bytes per 12x12=144 pixels
    TextureFormat.RG16: (8, 1),  # 2 bytes per pixel
    TextureFormat.R8: (16, 1),  # 1 byte per pixel
}


def get_padded_texture_size(width: int, height: int, block_width: int, block_height: int, texels_per_block: int):
    width = ceil_divide(width, block_width * GOB_X_TEXEL_COUNT) * block_width * GOB_X_TEXEL_COUNT
    height = (
        ceil_divide(height, block_height * GOB_Y_TEXEL_COUNT * texels_per_block)
        * block_height
        * GOB_Y_TEXEL_COUNT
        * texels_per_block
    )
    return width, height


def get_switch_gobs_per_block(platform_blob: List[int]) -> int:
    return 1 << int.from_bytes(platform_blob[8:12], "little")


def is_switch_swizzled(platform: Union[BuildTarget, int], platform_blob: Optional[List[int]]) -> bool:
    if platform != BuildTarget.Switch:
        return False
    if not platform_blob or len(platform_blob) < 12:
        return False
    gobs_per_block = get_switch_gobs_per_block(platform_blob)
    return gobs_per_block > 1



# based on: https://github.com/Razmoth/PGRStudio/blob/master/AssetStudio/PGR/PGR.cs
import re
from typing import Optional, Tuple, Union

from ..streams import EndianBinaryReader

try:
    from UnityPy import UnityPyBoost
except ImportError:
    UnityPyBoost = None

UNITY3D_SIGNATURE = b"#$unity3dchina!@"
DECRYPT_KEY: Optional[bytes] = None


def set_assetbundle_decrypt_key(key: Union[bytes, str]):
    if isinstance(key, str):
        key = key.encode("utf-8", "surrogateescape")
    if len(key) != 16:
        raise ValueError(f"AssetBundle Key length is wrong. It should be 16 bytes and now is {len(key)} bytes.")
    global DECRYPT_KEY
    DECRYPT_KEY = key


def read_vector(reader: EndianBinaryReader) -> Tuple[bytes, bytes]:
    data = reader.read_bytes(0x10)
    key = reader.read_bytes(0x10)
    reader.Position += 1

    return data, key


def decrypt_key(key: bytes, data: bytes, keybytes: bytes):
    from Crypto.Cipher import AES

    key = AES.new(keybytes, AES.MODE_ECB).encrypt(key)
    return bytes(x ^ y for x, y in zip(data, key))


def brute_force_key(
    fp: str,
    key_sig: bytes,
    data_sig: bytes,
    pattern: re.Pattern = re.compile(rb"(?=(\w{16}))"),
    verbose: bool = False,
):
    with open(fp, "rb") as f:
        data = f.read()

    matches = pattern.findall(data)
    for i, key in enumerate(matches):
        if verbose:
            print(f"Trying {i + 1}/{len(matches)} - {key}")
        signature = decrypt_key(key_sig, data_sig, key)
        if signature == UNITY3D_SIGNATURE:
            if verbose:
                print(f"Found key: {key}")
            return key
    return None


class ArchiveStorageDecryptor:
    unknown_1: int
    index: bytes
    substitute: bytes = bytes(0x10)

    def __init__(self, reader: EndianBinaryReader):
        self.unknown_1 = reader.read_u_int()

        # read vector data/key vectors
        self.data, self.key = read_vector(reader)
        self.data_sig, self.key_sig = read_vector(reader)

        if DECRYPT_KEY is None:
            raise LookupError(
                "\n".join(
                    [
                        "The BundleFile is encrypted, but no key was provided!",
                        "You can set the key via UnityPy.set_assetbundle_decrypt_key(key).",
                        "To try brute-forcing the key, use UnityPy.helpers.ArchiveStorageManager.brute_force_key(fp, key_sig, data_sig)",  # noqa: E501
                        f"with  key_sig = {self.key_sig}, data_sig = {self.data_sig},"
                        "and fp being the path to global-metadata.dat or a memory dump.",
                    ]
                )
            )

        signature = decrypt_key(self.key_sig, self.data_sig, DECRYPT_KEY)
        if signature != UNITY3D_SIGNATURE:
            raise Exception(f"Invalid signature {signature} != {UNITY3D_SIGNATURE}")

        data = decrypt_key(self.key, self.data, DECRYPT_KEY)
        data = bytes(nibble for byte in data for nibble in (byte >> 4, byte & 0xF))
        self.index = data[:0x10]
        self.substitute = bytes(data[0x10 + i * 4 + j] for j in range(4) for i in range(4))

    def decrypt_block(self, data: bytes, index: int):
        if UnityPyBoost:
            return UnityPyBoost.decrypt_block(self.index, self.substitute, data, index)

        offset = 0
        size = len(data)
        data = bytearray(data)
        view = memoryview(data)
        while offset < len(data):
            offset += self.decrypt(view[offset:], index, size - offset)
            index += 1
        return data

    def decrypt_byte(self, view: Union[bytearray, memoryview], offset: int, index: int):
        b = (
            self.substitute[((index >> 2) & 3) + 4]
            + self.substitute[index & 3]
            + self.substitute[((index >> 4) & 3) + 8]
            + self.substitute[(index % 256 >> 6) + 12]
        )
        view[offset] = ((self.index[view[offset] & 0xF] - b) & 0xF | 0x10 * (self.index[view[offset] >> 4] - b)) % 256
        b = view[offset]
        return b, offset + 1, index + 1

    def decrypt(self, data: Union[bytearray, memoryview], index: int, remaining: int):
        offset = 0

        curByte, offset, index = self.decrypt_byte(data, offset, index)
        byteHigh = curByte >> 4
        byteLow = curByte & 0xF

        if byteHigh == 0xF:
            b = 0xFF
            while b == 0xFF:
                b, offset, index = self.decrypt_byte(data, offset, index)
                byteHigh += b

        offset += byteHigh

        if offset < remaining:
            _, offset, index = self.decrypt_byte(data, offset, index)
            _, offset, index = self.decrypt_byte(data, offset, index)
            if byteLow == 0xF:
                b = 0xFF
                while b == 0xFF:
                    b, offset, index = self.decrypt_byte(data, offset, index)

        return offset

    # def encrypt(self, data: bytes):
    #     # TODO: patch BundleFile encryption flag to keep either 0x1000 or 0x400



from typing import TYPE_CHECKING, Any, List, Optional, Tuple

if TYPE_CHECKING:
    from ..classes.generated import PackedBitVector


def reshape(data: list, shape: Optional[Tuple[int, ...]] = None) -> List[Any]:
    if shape is None:
        return data
    if len(shape) == 1:
        m = shape[0]
        return [data[i : i + m] for i in range(0, len(data), m)]
    elif len(shape) == 2:
        m, n = shape
        return [[data[i + j : i + j + n] for j in range(0, m * n, n)] for i in range(0, len(data), m * n)]
    else:
        raise ValueError("Invalid shape")


def unpack_ints(
    packed: "PackedBitVector",
    start: int = 0,
    count: Optional[int] = None,
    shape: Optional[Tuple[int, ...]] = None,
) -> List[Any]:
    assert packed.m_BitSize is not None

    m_BitSize = packed.m_BitSize
    m_Data = packed.m_Data

    bitPos = m_BitSize * start
    indexPos = bitPos // 8
    bitPos %= 8

    if count is None:
        count = packed.m_NumItems

    # if m_BitSize <= 8:
    #     dtype = np.uint8
    # elif m_BitSize <= 16:
    #     dtype = np.uint16
    # elif m_BitSize <= 32:
    #     dtype = np.uint32
    # elif m_BitSize <= 64:
    #     dtype = np.uint64
    # else:
    #     raise ValueError("Invalid bit size")

    # data = np.zeros(packed.m_NumItems, dtype=dtype)
    data = [0] * count

    for i in range(count):
        bits = 0
        value = 0
        while bits < m_BitSize:
            value |= (m_Data[indexPos] >> bitPos) << bits
            num = min(m_BitSize - bits, 8 - bitPos)
            bitPos += num
            bits += num
            if bitPos == 8:
                indexPos += 1
                bitPos = 0
        data[i] = value & ((1 << m_BitSize) - 1)

    return reshape(data, shape)


def unpack_floats(
    packed: "PackedBitVector",
    start: int = 0,
    count: Optional[int] = None,
    shape: Optional[Tuple[int, ...]] = None,
) -> List[Any]:
    assert packed.m_BitSize is not None and packed.m_Range is not None and packed.m_Start is not None

    # avoid zero division of scale
    if packed.m_BitSize == 0:
        quantized = [packed.m_Start] * (packed.m_NumItems if count is None else count)
    else:
        # read as int and cast up to double to prevent loss of precision
        quantized_f64 = unpack_ints(packed, start, count)
        scale = packed.m_Range / ((1 << packed.m_BitSize) - 1)
        quantized = [x * scale + packed.m_Start for x in quantized_f64]

    return reshape(quantized, shape)


# def pack_ints(
#     data: npt.NDArray[np.uint], bitsize: Optional[int] = 0
# ) -> PackedBitVector:
#     # ensure that the data type is unsigned
#     assert "uint" in data.dtype.name

#     m_NumItems = data.size

#     maxi = data.max()
#     # Prevent overflow
#     if bitsize:
#         m_BitSize = bitsize
#     else:
#         m_BitSize = (32 if maxi == 0xFFFFFFFF else np.ceil(np.log2(maxi + 1))) % 256
#     m_Data = np.zeros((m_NumItems * m_BitSize + 7) // 8, dtype=np.uint8)

#     indexPos = 0
#     bitPos = 0
#     for x in data:
#         bits = 0
#         while bits < m_BitSize:
#             m_Data[indexPos] |= (x >> bits) << bitPos
#             num = min(m_BitSize - bits, 8 - bitPos)
#             bitPos += num
#             bits += num
#             if bitPos == 8:
#                 indexPos += 1
#                 bitPos = 0

#     return PackedBitVector(m_NumItems=m_NumItems, m_BitSize=m_BitSize, m_Data=m_Data)


# def pack_floats(
#     data: npt.NDArray[np.floating[Any]],
#     bitsize: Optional[int] = None,
# ) -> PackedBitVector:
#     min = data.min()
#     max = data.max()
#     range = max - min
#     data_f64 = data.astype(np.float64)
#     # rebase to 0
#     data_f64 -= min
#     # scale to [0, 1]
#     data_f64 /= range
#     # quantize to [0, 2^bit_size - 1]
#     bitsize = bitsize or max(data.itemsize, 32)
#     assert bitsize is not None

#     data_f64 *= (1 << bitsize) - 1
#     # pack the data
#     packed = pack_ints(data_f64.astype(np.uint32), bitsize)
#     packed.m_Start = min
#     packed.m_Range = range
#     return packed

__all__ = ("unpack_ints", "unpack_floats")



from __future__ import annotations

from typing import TYPE_CHECKING, Dict

import fmod_toolkit

from ..helpers.ResourceReader import get_resource_data

if TYPE_CHECKING:
    from ..classes import AudioClip


def extract_audioclip_samples(audio: AudioClip, convert_pcm_float: bool = True) -> Dict[str, bytes]:
    """extracts all the samples from an AudioClip
    :param audio: AudioClip
    :type audio: AudioClip
    :return: {filename : sample(bytes)}
    :rtype: dict
    """
    audio_data: bytes
    if audio.m_AudioData:
        audio_data = bytes(audio.m_AudioData)
    elif audio.m_Resource:
        assert audio.object_reader is not None, "AudioClip uses an external resource but object_reader is not set"
        resource = audio.m_Resource
        audio_data = get_resource_data(
            resource.m_Source,
            audio.object_reader.assets_file,
            resource.m_Offset,
            resource.m_Size,
        )
    else:
        raise ValueError("AudioClip with neither m_AudioData nor m_Resource")

    magic = memoryview(audio_data)[:8]
    if magic[:4] == b"OggS":
        return {f"{audio.m_Name}.ogg": audio_data}
    elif magic[:4] == b"RIFF":
        return {f"{audio.m_Name}.wav": audio_data}
    elif magic[4:8] == b"ftyp":
        return {f"{audio.m_Name}.m4a": audio_data}

    return fmod_toolkit.raw_to_wav(
        audio_data,
        audio.m_Name,
        audio.m_Channels or 2,
        audio.m_Frequency or 44100,
        convert_pcm_float=convert_pcm_float,
    )



from typing import BinaryIO, List, Optional, Union

from PIL.Image import Image

from UnityPy.classes.generated import GLTextureSettings, StreamingInfo, Texture

class Texture2D(Texture):
    image_data: bytes
    m_CompleteImageSize: int
    m_Height: int
    m_ImageCount: int
    m_IsReadable: bool
    m_LightmapFormat: int
    m_Name: str
    m_TextureDimension: int
    m_TextureFormat: int
    m_TextureSettings: GLTextureSettings
    m_Width: int
    m_ColorSpace: Optional[int] = None
    m_DownscaleFallback: Optional[bool] = None
    m_ForcedFallbackFormat: Optional[int] = None
    m_IgnoreMasterTextureLimit: Optional[bool] = None
    m_IgnoreMipmapLimit: Optional[bool] = None
    m_IsAlphaChannelOptional: Optional[bool] = None
    m_IsPreProcessed: Optional[bool] = None
    m_MipCount: Optional[int] = None
    m_MipMap: Optional[bool] = None
    m_MipmapLimitGroupName: Optional[str] = None
    m_MipsStripped: Optional[int] = None
    m_PlatformBlob: Optional[List[int]] = None
    m_ReadAllowed: Optional[bool] = None
    m_StreamData: Optional[StreamingInfo] = None
    m_StreamingMipmaps: Optional[bool] = None
    m_StreamingMipmapsPriority: Optional[int] = None

    @property
    def image(self) -> Image: ...
    def set_image(
        self,
        img: Union[Image, str, BinaryIO],
        target_format: Optional[int] = None,
        mipmap_count: int = 1,
    ) -> None: ...
    def get_image_data(self) -> bytes: ...



from .generated import *  # noqa: F403
from .legacy_patch import (
    AudioClip as AudioClip,
)
from .legacy_patch import (
    GameObject as GameObject,
)
from .legacy_patch import (
    Mesh as Mesh,
)
from .legacy_patch import (
    Renderer as Renderer,
)
from .legacy_patch import (
    Shader as Shader,
)
from .legacy_patch import (
    Sprite as Sprite,
)
from .legacy_patch import (
    Texture2D as Texture2D,
)
from .legacy_patch import (
    Texture2DArray as Texture2DArray,
)
from .Object import Object as Object
from .PPtr import PPtr as PPtr
from .UnknownObject import UnknownObject as UnknownObject



from enum import IntEnum


class TextureDimension(IntEnum):
    kTexDimUnknown = -1
    kTexDimNone = 0
    kTexDimAny = 1
    kTexDim2D = 2
    kTexDim3D = 3
    kTexDimCUBE = 4
    kTexDim2DArray = 5
    kTexDimCubeArray = 6
    kTexDimForce32Bit = 2147483647



from __future__ import annotations

import re
from struct import Struct
from threading import Lock
from typing import (
    TYPE_CHECKING,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)

from attrs import define, field

from ..streams.EndianBinaryReader import EndianBinaryReader
from ..streams.EndianBinaryWriter import EndianBinaryWriter

if TYPE_CHECKING:
    from .Tpk import UnityVersion

try:
    from ..UnityPyBoost import TypeTreeNode as TypeTreeNodeC  # type: ignore
except ImportError:

    @define(slots=True)
    class TypeTreeNodeC:
        m_Level: int
        m_Type: str
        m_Name: str
        m_ByteSize: int
        m_Version: int
        m_Children: List[TypeTreeNode] = field(factory=list)
        m_TypeFlags: Optional[int] = None
        m_VariableCount: Optional[int] = None
        m_Index: Optional[int] = None
        m_MetaFlag: Optional[int] = None
        m_RefTypeHash: Optional[int] = None
        _clean_name: str = field(init=False)

        def __attrs_post_init__(self):
            self._clean_name = clean_name(self.m_Name)

        def __repr__(self):
            return f"TypeTreeNode(m_Level={self.m_Level}, m_Type='{self.m_Type}', \
                m_Name='{self.m_Name}', m_MetaFlag={self.m_MetaFlag})"


TYPETREENODE_KEYS = [
    "m_Level",
    "m_Type",
    "m_Name",
    "m_ByteSize",
    "m_Version",
    "m_Children",
    "m_TypeFlags",
    "m_VariableCount",
    "m_Index",
    "m_MetaFlag",
    "m_RefTypeHash",
]

SYSTEM_GLOBAL_LOCK = Lock()
NAME_PEEK_NODE_CACHE: dict[Tuple[str, str, int], Union[Tuple[TypeTreeNode, str], None]] = {}


class TypeTreeNode(TypeTreeNodeC):
    def traverse(self) -> Iterator[TypeTreeNode]:
        stack: list[TypeTreeNode] = [self]
        while stack:
            node = stack.pop()
            yield node
            stack.extend(reversed(node.m_Children))

    @classmethod
    def parse(cls, reader: EndianBinaryReader, version: int) -> TypeTreeNode:
        # stack approach is way faster than recursion
        # using a fake root node to avoid special case for root node
        dummy_node = cls(-1, "", "", 0, 0, [])
        dummy_root = cls(-1, "", "", 0, 0, [dummy_node])

        stack: List[Tuple[TypeTreeNode, int]] = [(dummy_root, 1)]
        while stack:
            parent, count = stack[-1]
            if count == 1:
                stack.pop()
            else:
                stack[-1] = (parent, count - 1)

            node = cls(
                m_Level=parent.m_Level + 1,
                m_Type=reader.read_string_to_null(),
                m_Name=reader.read_string_to_null(),
                m_ByteSize=reader.read_int(),
                m_VariableCount=reader.read_int() if version == 2 else None,
                m_Index=reader.read_int() if version != 3 else None,
                m_TypeFlags=reader.read_int(),
                m_Version=reader.read_int(),
                m_MetaFlag=reader.read_int() if version != 3 else None,
            )
            parent.m_Children[-count] = node
            children_count = reader.read_int()
            if children_count > 0:
                node.m_Children = [dummy_node] * children_count
                stack.append((node, children_count))
        return dummy_root.m_Children[0]

    @classmethod
    def parse_blob(cls, reader: EndianBinaryReader, version: int) -> TypeTreeNode:
        node_count = reader.read_int()
        stringbuffer_size = reader.read_int()

        node_struct, keys = _get_blob_node_struct(reader.endian, version)
        struct_data = reader.read(node_struct.size * node_count)
        stringbuffer_reader = EndianBinaryReader(reader.read(stringbuffer_size), reader.endian)

        CommonString = get_common_strings()

        def read_string(reader: EndianBinaryReader, value: int) -> str:
            is_offset = (value & 0x80000000) == 0
            if is_offset:
                reader.Position = value
                return reader.read_string_to_null()

            offset = value & 0x7FFFFFFF
            return CommonString.get(offset, str(offset))

        fake_root: TypeTreeNode = cls(-1, "", "", 0, 0, [])
        stack: List[TypeTreeNode] = [fake_root]
        parent = fake_root
        prev = fake_root

        for raw_node in node_struct.iter_unpack(struct_data):
            node = cls(
                **dict(zip(keys[:3], raw_node[:3])),
                **dict(zip(keys[5:], raw_node[5:])),
                m_Type=read_string(stringbuffer_reader, raw_node[3]),
                m_Name=read_string(stringbuffer_reader, raw_node[4]),
            )

            if node.m_Level > prev.m_Level:
                stack.append(parent)
                parent = prev
            elif node.m_Level < prev.m_Level:
                while node.m_Level <= parent.m_Level:
                    parent = stack.pop()

            parent.m_Children.append(node)
            prev = node

        return fake_root.m_Children[0]

    @classmethod
    def from_list(cls, nodes: Union[List[Dict[str, Union[str, int]]], List[TypeTreeNode]]) -> TypeTreeNode:
        fake_root: TypeTreeNode = cls(-1, "", "", 0, 0, [])
        stack: List[TypeTreeNode] = [fake_root]
        parent = fake_root
        prev = fake_root

        # check if the nodes contain all required fields
        if isinstance(nodes[0], dict):
            if "m_Level" not in nodes[0] or "m_Type" not in nodes[0] or "m_Name" not in nodes[0]:
                raise ValueError("Nodes must contain at least m_Level, m_Type and m_Name")
            patch_dict = {}
            if "m_ByteSize" not in nodes[0]:
                patch_dict["m_ByteSize"] = 0
            if "m_Version" not in nodes[0]:
                patch_dict["m_Version"] = 0
            nodes = [cls(**node, **patch_dict) for node in nodes]  # type: ignore

        if TYPE_CHECKING:
            nodes = cast(List[TypeTreeNode], nodes)

        for node in nodes:
            if node.m_Level > prev.m_Level:
                stack.append(parent)
                parent = prev
            elif node.m_Level < prev.m_Level:
                while node.m_Level <= parent.m_Level:
                    parent = stack.pop()

            parent.m_Children.append(node)
            prev = node

        return fake_root.m_Children[0]

    def get_name_peek_node(self) -> Union[Tuple[TypeTreeNode, str], None]:
        global SYSTEM_GLOBAL_LOCK
        with SYSTEM_GLOBAL_LOCK:
            key = (self.m_Name, self.m_Type, self.m_Version)
            if key in NAME_PEEK_NODE_CACHE:
                return NAME_PEEK_NODE_CACHE[key]

            result: Union[Tuple[TypeTreeNode, str], None] = None
            for i, child in enumerate(self.m_Children):
                if child.m_Name in ("m_Name", "name"):
                    peek_node = TypeTreeNode(
                        self.m_Level,
                        self.m_Type,
                        self.m_Name,
                        self.m_ByteSize,
                        self.m_Version,
                        self.m_Children[: i + 1],
                    )
                    result = peek_node, child.m_Name
                    break
            NAME_PEEK_NODE_CACHE[key] = result
            return result

    def dump(self, writer: EndianBinaryWriter, version: int):
        stack: list[TypeTreeNode] = [self]
        while stack:
            node = stack.pop()

            writer.write_string_to_null(self.m_Type)
            writer.write_string_to_null(self.m_Name)
            writer.write_int(self.m_ByteSize)
            if version == 2:
                assert self.m_VariableCount is not None
                writer.write_int(self.m_VariableCount)
            if version != 3:
                assert self.m_Index is not None
                writer.write_int(self.m_Index)
            writer.write_int(self.m_TypeFlags or 0)
            writer.write_int(self.m_Version)
            if version != 3:
                assert self.m_MetaFlag is not None
                writer.write_int(self.m_MetaFlag)

            writer.write_int(len(self.m_Children))

            stack.extend(reversed(node.m_Children))

    def dump_blob(self, writer: EndianBinaryWriter, version: int):
        node_writer = EndianBinaryWriter(endian=writer.endian)
        string_writer = EndianBinaryWriter()

        # string buffer setup
        CommonStringOffsetMap = {string: offset for offset, string in get_common_strings().items()}

        string_offsets: dict[str, int] = {}

        def write_string(string: str) -> int:
            offset = string_offsets.get(string)
            if offset is None:
                common_offset = CommonStringOffsetMap.get(string)
                if common_offset:
                    offset = common_offset | 0x80000000
                else:
                    offset = string_writer.Position
                    string_writer.write_string_to_null(string)
                string_offsets[string] = offset
            return offset

        # node buffer setup
        node_struct, keys = _get_blob_node_struct(writer.endian, version)

        def write_node(node: TypeTreeNode):
            node_writer.write(
                node_struct.pack(
                    *[getattr(node, key) for key in keys[:3]],
                    write_string(node.m_Type),
                    write_string(node.m_Name),
                    *[getattr(node, key) for key in keys[5:]],
                )
            )

        # write nodes
        node_count = len([write_node(node) for node in self.traverse()])

        # write blob
        writer.write_int(node_count)
        writer.write_int(string_writer.Position)
        writer.write(node_writer.bytes)
        writer.write(string_writer.bytes)

    def dump_structure(self, indent: str = "  ") -> str:
        # dump structure similar to https://github.com/AssetRipper/TypeTreeDumps/blob/main/StructsDump
        sb = [
            f"{indent}{self.m_Type} {self.m_Name} // ByteSize{{{self.m_ByteSize:X}}}, Index{{{self.m_Index}}}, \
                Version{{{self.m_Version}}}, TypeFlags{{{self.m_TypeFlags}}}, MetaFlag{{{self.m_MetaFlag}}}"
        ]
        for child in self.m_Children:
            sb.append(child.dump_structure(indent + "  "))
        return "\n".join(sb)

    def to_dict(self) -> dict:
        return {
            key: value for key, value in ((key, getattr(self, key)) for key in TYPETREENODE_KEYS) if value is not None
        }

    def to_dict_list(self) -> List[dict]:
        return [
            self.to_dict(),
            *(item for child in self.m_Children for item in child.to_dict_list()),
        ]

    def __eq__(self, other: TypeTreeNode) -> bool:  # type: ignore
        return self.to_dict() == other.to_dict() and self.m_Children == other.m_Children


COMMONSTRING_CACHE: Dict[Optional[UnityVersion], Dict[int, str]] = {}


def get_common_strings(version: Optional[UnityVersion] = None) -> Dict[int, str]:
    if version in COMMONSTRING_CACHE:
        return COMMONSTRING_CACHE[version]

    from .Tpk import TPKTYPETREE

    tree = TPKTYPETREE
    common_string = tree.CommonString
    strings = common_string.GetStrings(tree.StringBuffer)
    if version:
        count = common_string.GetCount(version)
        strings = strings[:count]

    ret: Dict[int, str] = {}
    offset = 0
    for string in strings:
        ret[offset] = string
        offset += len(string) + 1

    COMMONSTRING_CACHE[version] = ret
    return ret


def _get_blob_node_struct(endian: str, version: int) -> tuple[Struct, list[str]]:
    struct_type = f"{endian}hBBIIiii"
    keys = [
        "m_Version",
        "m_Level",
        "m_TypeFlags",
        "m_TypeStrOffset",
        "m_NameStrOffset",
        "m_ByteSize",
        "m_Index",
        "m_MetaFlag",
    ]
    if version >= 19:
        struct_type += "Q"
        keys.append("m_RefTypeHash")

    return Struct(struct_type), keys


def clean_name(name: str) -> str:
    # keep in sync with TypeTreeHelper.cpp
    if len(name) == 0:
        return name
    if name.startswith("(int&)"):
        name = name[6:]
    if name.endswith("?"):
        name = name[:-1]
    name = re.sub(r"[ \.:\-\[\]]", "_", name)
    if name in ["pass", "from"]:
        name += "_"
    if name[0].isdigit():
        name = f"x{name}"
    return name


__all__ = (
    "TypeTreeNode",
    "get_common_strings",
    "clean_name",
)



from enum import IntEnum


class VertexChannelFormat(IntEnum):
    kChannelFormatFloat = 0
    kChannelFormatFloat16 = 1
    kChannelFormatColor = 2
    kChannelFormatByte = 3
    kChannelFormatUInt32 = 4


class VertexFormat2017(IntEnum):
    kVertexFormatFloat = 0
    kVertexFormatFloat16 = 1
    kVertexFormatColor = 2
    kVertexFormatUNorm8 = 3
    kVertexFormatSNorm8 = 4
    kVertexFormatUNorm16 = 5
    kVertexFormatSNorm16 = 6
    kVertexFormatUInt8 = 7
    kVertexFormatSInt8 = 8
    kVertexFormatUInt16 = 9
    kVertexFormatSInt16 = 10
    kVertexFormatUInt32 = 11
    kVertexFormatSInt32 = 12


class VertexFormat(IntEnum):
    kVertexFormatFloat = 0
    kVertexFormatFloat16 = 1
    kVertexFormatUNorm8 = 2
    kVertexFormatSNorm8 = 3
    kVertexFormatUNorm16 = 4
    kVertexFormatSNorm16 = 5
    kVertexFormatUInt8 = 6
    kVertexFormatSInt8 = 7
    kVertexFormatUInt16 = 8
    kVertexFormatSInt16 = 9
    kVertexFormatUInt32 = 10
    kVertexFormatSInt32 = 11


VERTEX_CHANNEL_FORMAT_STRUCT_TYPE_MAP = {
    VertexChannelFormat.kChannelFormatFloat: "f",
    VertexChannelFormat.kChannelFormatFloat16: "e",
    VertexChannelFormat.kChannelFormatColor: "B",
    VertexChannelFormat.kChannelFormatByte: "B",
    VertexChannelFormat.kChannelFormatUInt32: "I",
}

VERTEX_FORMAT_2017_STRUCT_TYPE_MAP = {
    VertexFormat2017.kVertexFormatFloat: "f",
    VertexFormat2017.kVertexFormatFloat16: "e",
    VertexFormat2017.kVertexFormatColor: "B",
    VertexFormat2017.kVertexFormatUNorm8: "B",
    VertexFormat2017.kVertexFormatSNorm8: "b",
    VertexFormat2017.kVertexFormatUNorm16: "H",
    VertexFormat2017.kVertexFormatSNorm16: "h",
    VertexFormat2017.kVertexFormatUInt8: "B",
    VertexFormat2017.kVertexFormatSInt8: "b",
    VertexFormat2017.kVertexFormatUInt16: "H",
    VertexFormat2017.kVertexFormatSInt16: "h",
    VertexFormat2017.kVertexFormatUInt32: "I",
    VertexFormat2017.kVertexFormatSInt32: "i",
}

VERTEX_FORMAT_STRUCT_TYPE_MAP = {
    VertexFormat.kVertexFormatFloat: "f",
    VertexFormat.kVertexFormatFloat16: "e",
    VertexFormat.kVertexFormatUNorm8: "B",
    VertexFormat.kVertexFormatSNorm8: "b",
    VertexFormat.kVertexFormatUNorm16: "H",
    VertexFormat.kVertexFormatSNorm16: "h",
    VertexFormat.kVertexFormatUInt8: "B",
    VertexFormat.kVertexFormatSInt8: "b",
    VertexFormat.kVertexFormatUInt16: "H",
    VertexFormat.kVertexFormatSInt16: "h",
    VertexFormat.kVertexFormatUInt32: "I",
    VertexFormat.kVertexFormatSInt32: "i",
}



from __future__ import annotations

from collections import namedtuple
from os.path import basename
from typing import TYPE_CHECKING, Dict, Optional

from ..helpers import ImportHelper
from ..streams import EndianBinaryReader, EndianBinaryWriter

if TYPE_CHECKING:
    from ..environment import Environment

DirectoryInfo = namedtuple("DirectoryInfo", "path offset size")


class File:
    name: str
    files: Dict[str, File]
    environment: Environment
    cab_file: str
    is_changed: bool
    signature: str
    packer: str
    is_dependency: bool
    parent: Optional[File]

    def __init__(
        self,
        parent: Optional[File] = None,
        name: Optional[str] = None,
        is_dependency: bool = False,
    ):
        self.files = {}
        self.is_changed = False
        self.cab_file = "CAB-UnityPy_Mod.resS"
        self.parent = parent
        self.environment = getattr(parent, "environment", parent) if parent else None
        self.name = basename(name) if isinstance(name, str) else ""
        self.is_dependency = is_dependency

    def get_assets(self):
        if isinstance(self, SerializedFile.SerializedFile):
            return self

        for f in self.files.values():
            if isinstance(f, (BundleFile.BundleFile, WebFile.WebFile)):
                for asset in f.get_assets():
                    yield asset
            elif isinstance(f, SerializedFile.SerializedFile):
                yield f

    def get_filtered_objects(self, obj_types: Optional[list] = None):
        if obj_types is None or len(obj_types) == 0:
            return self.get_objects()
        for f in self.files.values():
            if isinstance(f, (BundleFile.BundleFile, WebFile.WebFile)):
                for obj in f.objects:
                    if obj.type in obj_types:
                        yield obj
            elif isinstance(f, SerializedFile.SerializedFile):
                for obj in f.objects.values():
                    if obj.type in obj_types:
                        yield obj

    def get_objects(self):
        for f in self.files.values():
            if isinstance(f, (BundleFile.BundleFile, WebFile.WebFile)):
                for obj in f.objects:
                    yield obj
            elif isinstance(f, SerializedFile.SerializedFile):
                for obj in f.objects.values():
                    yield obj
            elif isinstance(f, ObjectReader.ObjectReader):
                yield f

    def read_files(self, reader: EndianBinaryReader, files: list):
        # read file data and convert it
        for node in files:
            reader.Position = node.offset
            name = node.path
            node_reader = EndianBinaryReader(reader.read(node.size), offset=(reader.BaseOffset + node.offset))
            f = ImportHelper.parse_file(node_reader, self, name, is_dependency=self.is_dependency)

            if isinstance(f, (EndianBinaryReader, SerializedFile.SerializedFile)):
                if self.environment:
                    self.environment.register_cab(name, f)

            # required for BundleFiles
            f.flags = getattr(node, "flags", 0)
            self.files[name] = f

    def get_writeable_cab(self, name: Optional[str] = None):
        """
        Creates a new cab file in the bundle that contains the given data.
        This is useful for asset types that use resource files.
        """

        if not name:
            name = self.cab_file

        if not name:
            return None

        if name in self.files:
            if isinstance(self.files[name], EndianBinaryWriter):
                return self.files[name]
            else:
                raise ValueError("This cab already exists and isn't an EndianBinaryWriter")

        writer = EndianBinaryWriter()
        # try to find another resource file to copy the flags from
        for fname, f in self.files.items():
            if fname.endswith(".resS"):
                writer.flags = f.flags
                writer.endian = f.endian
                break
        else:
            writer.flags = 0
        writer.name = name
        self.files[name] = writer
        return writer

    @property
    def container(self):
        return {path: obj for f in self.files.values() if isinstance(f, File) for path, obj in f.container.items()}

    def get(self, key, default=None):
        return getattr(self, key, default)

    def keys(self):
        return self.files.keys()

    def items(self):
        return self.files.items()

    def values(self):
        return self.files.values()

    def __getitem__(self, item):
        return self.files[item]

    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    def mark_changed(self):
        if isinstance(self.parent, File):
            self.parent.mark_changed()
        self.is_changed = True


# recursive import requires the import down here
from . import BundleFile, ObjectReader, SerializedFile, WebFile  # noqa: E402



from ..generated import Shader


def _Shader_export(self: Shader) -> str:
    from ...export.ShaderConverter import export_shader

    return export_shader(self)


Shader.export = _Shader_export

__all__ = ("Shader",)



from __future__ import annotations

import re
from enum import IntEnum
from typing import Optional, Tuple, Union

VersionPattern = re.compile(r"^(?P<major>\d+)\.(?P<minor>\d+)\.(?P<build>\d+)(?P<type_str>.+?)?(?P<type_number>\d+)?$")


class UnityVersionType(IntEnum):
    a = 0  # Alpha
    b = 1  # Beta
    c = 2  # China
    f = 3  # Final
    p = 4  # Patch
    x = 5  # Experimental
    u = 255  # Unknown


class UnityVersion(int):
    # https://github.com/AssetRipper/VersionUtilities/blob/master/VersionUtilities/UnityVersion.cs
    _type_str: Optional[str]

    @property
    def major(self):
        return (self >> 48) & 0xFFFF

    @property
    def minor(self):
        return (self >> 32) & 0xFFFF

    @property
    def build(self):
        return (self >> 16) & 0xFFFF

    @property
    def type(self):
        return UnityVersionType((self >> 8) & 0xFF)

    @property
    def type_str(self):
        return getattr(self, "_type_str", self.type.name)

    @property
    def type_number(self):
        return self & 0xFF

    @classmethod
    def from_list(
        cls, major: int = 0, minor: int = 0, build: int = 0, type: int = UnityVersionType.f, type_number: int = 0
    ):
        return cls((major << 48) | (minor << 32) | (build << 16) | (type << 8) | type_number)

    @classmethod
    def from_str(cls, version: str):
        # formats:
        #   old: 5.0.0, <major>.<minor>.<build>
        #   new: 2018.1.1f2 <major>.<minor>.<build><type_str><type_number>
        match = VersionPattern.match(version)
        if not match:
            raise ValueError(f"Invalid version string: {version}")
        major = int(match.group("major"))
        minor = int(match.group("minor"))
        build = int(match.group("build"))
        type_str = match.group("type_str")
        type_number = int(match.group("type_number") or 0)

        if type_str is None:
            return cls.from_list(major, minor, build)

        type = getattr(UnityVersionType, type_str.lower(), UnityVersionType.u)
        obj = cls.from_list(major, minor, build, type, type_number)
        if type is UnityVersionType.u:
            obj._type_str = type_str
        return obj

    def __repr__(self) -> str:
        return f"UnityVersion {self.major}.{self.minor}{self.type_str}{self.type_number}"

    def __getitem__(self, idx: int) -> int:
        if idx == 0:
            return self.major
        elif idx == 1:
            return self.minor
        elif idx == 2:
            return self.build
        elif idx == 3:
            return self.type.value
        elif idx == 4:
            return self.type_number
        raise IndexError("Invalid UnityVersion index")

    def as_tuple(self) -> Tuple[int, int, int, int, int]:
        return (self.major, self.minor, self.build, self.type.value, self.type_number)

    def __eq__(self, other: Union[int, UnityVersion, Tuple[int, ...]]) -> bool:
        if isinstance(other, int):
            return super().__eq__(other)
        elif isinstance(other, tuple):
            return self.as_tuple() == other
        raise NotImplementedError("Unsupported comparison")

    def __ne__(self, other: Union[int, UnityVersion, Tuple[int, ...]]) -> bool:
        return not self.__eq__(other)

    def __lt__(self, other: Union[int, UnityVersion, Tuple[int, ...]]) -> bool:
        if isinstance(other, int):
            return super().__lt__(other)
        elif isinstance(other, tuple):
            return self.as_tuple() < other
        raise NotImplementedError("Unsupported comparison")

    def __le__(self, other: Union[int, UnityVersion, Tuple[int, ...]]) -> bool:
        return self.__lt__(other) or self.__eq__(other)

    def __gt__(self, other: Union[int, UnityVersion, Tuple[int, ...]]) -> bool:
        if isinstance(other, int):
            return super().__gt__(other)
        elif isinstance(other, tuple):
            return self.as_tuple() > other
        raise NotImplementedError("Unsupported comparison")

    def __ge__(self, other: Union[int, UnityVersion, Tuple[int, ...]]) -> bool:
        return self.__gt__(other) or self.__eq__(other)

    def __hash__(self) -> int:
        return super().__hash__()


__all__ = ["UnityVersion", "UnityVersionType"]



import ntpath
from typing import TYPE_CHECKING

from ..streams import EndianBinaryReader

if TYPE_CHECKING:
    from ..files.SerializedFile import SerializedFile


def get_resource_data(res_path: str, assets_file: "SerializedFile", offset: int, size: int):
    basename = ntpath.basename(res_path)
    name, ext = ntpath.splitext(basename)
    possible_names = [
        basename,
        f"{name}.resource",
        f"{name}.assets.resS",
        f"{name}.resS",
    ]
    environment = assets_file.environment
    reader = None
    for possible_name in possible_names:
        reader = environment.get_cab(possible_name)
        if reader:
            break
    if not reader:
        assets_file.load_dependencies(possible_names)
        for possible_name in possible_names:
            reader = environment.get_cab(possible_name)
            if reader:
                break
        if not reader:
            raise FileNotFoundError(f"Resource file {basename} not found")
    return _get_resource_data(reader, offset, size)


def _get_resource_data(reader: EndianBinaryReader, offset: int, size: int):
    reader.Position = offset
    return reader.read_bytes(size)



from ..generated import Mesh


def _Mesh_export(self: Mesh, format: str = "obj"):
    from ...export.MeshExporter import export_mesh

    return export_mesh(self, format)


Mesh.export = _Mesh_export


__all__ = ("Mesh",)



from __future__ import annotations

from enum import IntEnum, IntFlag
from importlib.resources import open_binary
from io import BytesIO
from struct import Struct
from typing import Any, Dict, List, Optional, Tuple, TypeVar

from .TypeTreeHelper import TypeTreeNode
from .UnityVersion import UnityVersion

T = TypeVar("T")

TPKTYPETREE: TpkTypeTreeBlob = None  # pyright: ignore[reportAssignmentType]
CLASSES_CACHE: Dict[Tuple[int, UnityVersion], TypeTreeNode] = {}
NODES_CACHE: Dict[TpkUnityClass, TypeTreeNode] = {}


def init():
    with open_binary("UnityPy.resources", "uncompressed.tpk") as f:
        data = f.read()

    global TPKTYPETREE
    with BytesIO(data) as stream:
        blob = TpkFile(stream).GetDataBlob()
        assert isinstance(blob, TpkTypeTreeBlob)
        TPKTYPETREE = blob


def get_typetree_node(class_id: int, version: UnityVersion):
    global CLASSES_CACHE
    key = (class_id, version)
    cached = CLASSES_CACHE.get(key)
    if cached:
        return cached

    class_info = TPKTYPETREE.ClassInformation[class_id].getVersionedClass(version)
    if class_info is None:
        raise ValueError("Could not find class info for class id {}".format(class_id))

    node = generate_node(class_info)
    CLASSES_CACHE[key] = node
    return node


def generate_node(class_info: TpkUnityClass) -> TypeTreeNode:
    global NODES_CACHE
    cached = NODES_CACHE.get(class_info)
    if cached:
        return cached

    assert class_info.ReleaseRootNode is not None, "Class {} has no ReleaseRootNode".format(class_info)

    nodes = []
    NODES = TPKTYPETREE.NodeBuffer.Nodes
    stack = [(class_info.ReleaseRootNode, 0)]
    index = 0
    while stack:
        node_id, level = stack.pop(0)
        node: TpkUnityNode = NODES[node_id]
        nodes.append(
            TypeTreeNode(
                m_ByteSize=node.ByteSize,
                m_Index=index,
                m_Version=node.Version,
                m_MetaFlag=node.MetaFlag,
                m_Level=level,
                m_Type=TPKTYPETREE.StringBuffer.Strings[node.TypeName],
                m_Name=TPKTYPETREE.StringBuffer.Strings[node.Name],
            )
        )
        stack = [(node_id, level + 1) for node_id in node.SubNodes] + stack
        index += 1
    result = TypeTreeNode.from_list(nodes)
    NODES_CACHE[class_info] = result
    return result


######################################################################################
#
#   Enums
#
######################################################################################


class TpkCompressionType(IntEnum):
    NONE = 0
    Lz4 = 1
    Lzma = 2
    Brotli = 3


class TpkDataType(IntEnum):
    TypeTreeInformation = 0
    Collection = 1
    FileSystem = 2
    Json = 3
    ReferenceAssemblies = 4
    EngineAssets = 5

    def ToBlob(self, stream):
        if self.value == TpkDataType.TypeTreeInformation:
            return TpkTypeTreeBlob(stream)
        elif self.value == TpkDataType.Collection:
            return TpkCollectionBlob(stream)
        elif self.value == TpkDataType.FileSystem:
            return TpkFileSystemBlob(stream)
        elif self.value == TpkDataType.Json:
            return TpkJsonBlob(stream)
        else:
            raise Exception("Unimplemented TpkDataType -> Blob conversion")


class TpkUnityClassFlags(IntFlag):
    NONE = 0
    IsAbstract = 1
    IsSealed = 2
    IsEditorOnly = 4
    IsReleaseOnly = 8
    IsStripped = 16
    Reserved = 32
    HasEditorRootNode = 64
    HasReleaseRootNode = 128


######################################################################################
#
#   Main Class
#
######################################################################################


class TpkFile:
    Struct = Struct("<IbbbbIII")
    TpkMagicBytes: int = 0x2A4B5054  # b"TPK*"
    TpkVersionNumber: int = 1
    CompressionType: TpkCompressionType
    DataType: TpkDataType
    CompressedSize: int
    UncompressedSize: int
    CompressedBytes: bytes

    def __init__(self, stream: BytesIO):
        (
            magic,
            versionNumber,
            compressionType,
            dataType,
            _,
            _,
            self.CompressedSize,
            self.UncompressedSize,
        ) = TpkFile.Struct.unpack(stream.read(TpkFile.Struct.size))
        if magic != TpkFile.TpkMagicBytes:
            raise Exception("Invalid TPK magic bytes")
        if versionNumber != TpkFile.TpkVersionNumber:
            raise Exception("Invalid TPK version number")
        self.CompressionType = TpkCompressionType(compressionType)
        self.DataType = TpkDataType(dataType)
        self.CompressedBytes = stream.read(self.CompressedSize)
        if len(self.CompressedBytes) != self.CompressedSize:
            raise Exception("Invalid compressed size")

    def GetDataBlob(self) -> TpkDataBlob:
        decompressed: bytes
        if self.CompressionType == TpkCompressionType.NONE:
            decompressed = self.CompressedBytes

        elif self.CompressionType == TpkCompressionType.Lz4:
            import lz4.block

            decompressed = lz4.block.decompress(self.CompressedBytes, self.UncompressedSize)

        elif self.CompressionType == TpkCompressionType.Lzma:
            import lzma

            decompressed = lzma.decompress(self.CompressedBytes)

        elif self.CompressionType == TpkCompressionType.Brotli:
            import brotli

            decompressed = brotli.decompress(self.CompressedBytes)

        else:
            raise Exception("Invalid compression type")

        return self.DataType.ToBlob(BytesIO(decompressed))


######################################################################################
#
#   Blobs
#
######################################################################################


class TpkDataBlob:
    __slots__ = ("DataType",)
    DataType: TpkDataType

    def __init__(self, stream: BytesIO) -> None:
        raise NotImplementedError("TpkDataBlob is an abstract class")


class TpkTypeTreeBlob(TpkDataBlob):
    __slots__ = (
        "CreationTime",
        "Versions",
        "ClassInformation",
        "CommonString",
        "NodeBuffer",
        "StringBuffer",
    )
    CreationTime: int
    Versions: List[UnityVersion]
    ClassInformation: Dict[int, TpkClassInformation]  # List[TpkClassInformation]
    CommonString: TpkCommonString
    NodeBuffer: TpkUnityNodeBuffer
    StringBuffer: TpkStringBuffer
    DataType: TpkDataType = TpkDataType.TypeTreeInformation

    def __init__(self, stream: BytesIO) -> None:
        (self.CreationTime,) = INT64.unpack(stream.read(INT64.size))
        (versionCount,) = INT32.unpack(stream.read(INT32.size))
        self.Versions = [read_version(stream) for _ in range(versionCount)]
        (classCount,) = INT32.unpack(stream.read(INT32.size))
        self.ClassInformation = {x.ID: x for x in (TpkClassInformation(stream) for _ in range(classCount))}
        self.CommonString = TpkCommonString(stream)
        self.NodeBuffer = TpkUnityNodeBuffer(stream)
        self.StringBuffer = TpkStringBuffer(stream)


class TpkCollectionBlob(TpkDataBlob):
    __slots__ = "Blobs"
    Blobs: List[Tuple[str, TpkDataBlob]]

    def __init__(self, stream: BytesIO) -> None:
        (count,) = INT32.unpack(stream.read(INT32.size))
        self.Blobs = [
            # relativePath, data
            (
                read_string(stream),
                TpkDataType(BYTE.unpack(stream.read(1))[0]).ToBlob(stream),
            )
            for _ in range(count)
        ]


class TpkFileSystemBlob(TpkDataBlob):
    __slots__ = ("Files",)
    # TODO: check if dict might be better
    Files: List[Tuple[str, bytes]]

    def __init__(self, stream: BytesIO) -> None:
        (count,) = INT32.unpack(stream.read(INT32.size))
        self.Files = [
            # relativePath, data
            (read_string(stream), read_data(stream))
            for _ in range(count)
        ]


class TpkJsonBlob(TpkDataBlob):
    __slots__ = "Text"
    Text: str
    DataType = TpkDataType.Json

    def __init__(self, stream: BytesIO) -> None:
        self.Text = read_string(stream)


######################################################################################
#
#   Unity
#
######################################################################################


class TpkUnityClass:
    __slots__ = ("Name", "Base", "Flags", "EditorRootNode", "ReleaseRootNode")
    Struct = Struct("<HHb")
    Name: int
    Base: int
    Flags: TpkUnityClassFlags
    EditorRootNode: Optional[int]
    ReleaseRootNode: Optional[int]

    def __init__(self, stream: BytesIO) -> None:
        self.Name, self.Base, Flags = TpkUnityClass.Struct.unpack(stream.read(TpkUnityClass.Struct.size))
        self.Flags = TpkUnityClassFlags(Flags)
        self.EditorRootNode = self.ReleaseRootNode = None
        if self.Flags & TpkUnityClassFlags.HasEditorRootNode:
            (self.EditorRootNode,) = UINT16.unpack(stream.read(UINT16.size))
        if self.Flags & TpkUnityClassFlags.HasReleaseRootNode:
            (self.ReleaseRootNode,) = UINT16.unpack(stream.read(UINT16.size))

    def to_dict(self) -> Dict[str, Any]:
        return {
            "Name": self.Name,
            "Base": self.Base,
            "Flags": self.Flags,
            "EditorRootNode": self.EditorRootNode,
            "ReleaseRootNode": self.ReleaseRootNode,
        }

    def __eq__(self, other: TpkUnityClass) -> bool:
        return self.to_dict() == other.to_dict()

    def __hash__(self) -> int:
        return hash(
            (
                self.Name,
                self.Base,
                self.Flags,
                self.EditorRootNode,
                self.ReleaseRootNode,
            )
        )


class TpkClassInformation:
    __slots__ = ("ID", "Classes")
    ID: int
    # TODO - might want to use dict
    Classes: List[Tuple[UnityVersion, Optional[TpkUnityClass]]]

    def __init__(self, stream: BytesIO) -> None:
        (self.ID,) = INT32.unpack(stream.read(INT32.size))
        (count,) = INT32.unpack(stream.read(INT32.size))
        self.Classes = [
            (
                read_version(stream),
                TpkUnityClass(stream) if stream.read(1)[0] else None,
            )
            for _ in range(count)
        ]

    def getVersionedClass(self, version: UnityVersion) -> Optional[TpkUnityClass]:
        return get_item_for_version(version, self.Classes)


class TpkUnityNodeBuffer:
    Nodes: List[TpkUnityNode]

    def __init__(self, stream: BytesIO) -> None:
        (count,) = INT32.unpack(stream.read(INT32.size))
        self.Nodes = [TpkUnityNode(stream) for _ in range(count)]

    def __getitem__(self, index: int) -> TpkUnityNode:
        return self.Nodes[index]


class TpkUnityNode:
    __slots__ = (
        "TypeName",
        "Name",
        "ByteSize",
        "Version",
        "TypeFlags",
        "MetaFlag",
        "SubNodes",
    )
    Struct = Struct("<HHihbIH")
    TypeName: int
    Name: int
    ByteSize: int
    Version: int
    TypeFlags: int
    MetaFlag: int
    SubNodes: List[int]

    def __init__(self, stream: BytesIO) -> None:
        (
            self.TypeName,
            self.Name,
            self.ByteSize,
            self.Version,
            self.TypeFlags,
            self.MetaFlag,
            count,
        ) = TpkUnityNode.Struct.unpack(stream.read(TpkUnityNode.Struct.size))

        SubNodeStruct = Struct(f"<{count}H")
        self.SubNodes = list(SubNodeStruct.unpack(stream.read(SubNodeStruct.size)))

    def __eq__(self, other: TpkUnityNode) -> bool:
        return self.__dict__ == other.__dict__

    def __hash__(self) -> int:
        # TODO
        return hash(self.__dict__)


######################################################################################
#
#   Strings
#
######################################################################################


class TpkStringBuffer:
    __slots__ = "Strings"
    Strings: List[str]

    def __init__(self, stream: BytesIO) -> None:
        self.Strings = [read_string(stream) for _ in range(INT32.unpack(stream.read(INT32.size))[0])]

    @property
    def Count(self) -> int:
        return len(self.Strings)


class TpkCommonString:
    __slots__ = ("VersionInformation", "StringBufferIndices")
    VersionInformation: List[Tuple[UnityVersion, int]]
    StringBufferIndices: Tuple[int, ...]

    def __init__(self, stream: BytesIO) -> None:
        (versionCount,) = INT32.unpack(stream.read(INT32.size))
        self.VersionInformation = [(read_version(stream), stream.read(1)[0]) for _ in range(versionCount)]
        (indicesCount,) = INT32.unpack(stream.read(INT32.size))
        indicesStruct = Struct(f"<{indicesCount}H")
        self.StringBufferIndices = indicesStruct.unpack(stream.read(indicesStruct.size))

    def GetStrings(self, buffer: TpkStringBuffer) -> List[str]:
        return [buffer.Strings[i] for i in self.StringBufferIndices]

    def GetCount(self, exactVersion: UnityVersion) -> int:
        return get_item_for_version(exactVersion, self.VersionInformation)


######################################################################################
#
# helper functions
#
######################################################################################

BYTE = Struct("b")
UINT16 = Struct("<H")
INT32 = Struct("<i")
INT64 = Struct("<q")
UINT64 = Struct("<Q")


def read_string(stream: BytesIO) -> str:
    # varint
    shift = 0
    length = 0
    while True:
        (i,) = stream.read(1)
        length |= (i & 0x7F) << shift
        shift += 7
        if not (i & 0x80):
            break
    # string
    return stream.read(length).decode("utf-8")


def read_data(stream: BytesIO) -> bytes:
    return stream.read(INT32.unpack(stream.read(INT32.size))[0])


def read_version(stream: BytesIO) -> UnityVersion:
    return UnityVersion(UINT64.unpack(stream.read(UINT64.size))[0])


def get_item_for_version(exactVersion: UnityVersion, items: List[Tuple[UnityVersion, T]]) -> T:
    ret = None
    for version, item in items:
        if exactVersion >= version:
            ret = item
        else:
            break
    if ret:
        return ret
    raise ValueError("Could not find exact version")


init()



from __future__ import annotations

from abc import ABC, ABCMeta
from typing import TYPE_CHECKING, Any, Dict, Optional

if TYPE_CHECKING:
    from ..files.ObjectReader import ObjectReader
    from ..files.SerializedFile import SerializedFile


class Object(ABC, metaclass=ABCMeta):
    object_reader: Optional[ObjectReader] = None

    def __init__(self, **kwargs: Dict[str, Any]) -> None:
        self.__dict__.update(**kwargs)

    def set_object_reader(self, object_info: ObjectReader[Any]):
        self.object_reader = object_info

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"

    @property
    def assets_file(self) -> Optional[SerializedFile]:
        if self.object_reader:
            return self.object_reader.assets_file
        return None

    def save(self) -> None:
        if self.object_reader is None:
            raise ValueError("ObjectReader not set")

        self.object_reader.save_typetree(self)


__all__ = [
    "Object",
]



from typing import Optional

from ..helpers.TypeTreeNode import TypeTreeNode
from .Object import Object


class UnknownObject(Object):
    """An object of unknown type that showed up during typetree parsing."""

    __node__: Optional[TypeTreeNode]

    def __init__(self, __node__: Optional[TypeTreeNode] = None, **kwargs):
        self.__node__ = __node__
        self.__dict__.update(**kwargs)

    def get_type(self):
        return self.__node__.m_Type if self.__node__ else None

    def __repr__(self) -> str:
        def format_value(v):
            vstr = repr(v)
            if len(vstr) > 100:
                return vstr[:97] + "..."
            return vstr

        inner_str = ", ".join(f"{k}={format_value(v)}" for k, v in self.__dict__.items() if k != "__node__")

        return f"<UnknownObject<{self.get_type()}> {inner_str}>"


__all__ = [
    "UnknownObject",
]



import warnings

from .exceptions import UnityVersionFallbackError, UnityVersionFallbackWarning

FALLBACK_UNITY_VERSION = None
"""The Unity version to use when no version is defined
   by the SerializedFile or its BundleFile.

   You may manually configure this value to a version string, e.g. `2.5.0f5`.
"""

SERIALIZED_FILE_PARSE_TYPETREE = True
"""Determines if the typetree structures for the Object types will be parsed.

   Disabling this will reduce the load time by a lot (half of the time is spend on parsing the typetrees),
   but it will also prevent saving an edited file.
"""


# WARNINGS CONTROL
warnings.simplefilter("once", UnityVersionFallbackWarning)


# GET FUNCTIONS
def get_fallback_version():
    global FALLBACK_UNITY_VERSION

    if not isinstance(FALLBACK_UNITY_VERSION, str):
        raise UnityVersionFallbackError(
            "No valid Unity version found, and the fallback version is not correctly configured. "
            + "Please explicitly set the value of UnityPy.config.FALLBACK_UNITY_VERSION."
        )

    warnings.warn(
        f"No valid Unity version found, defaulting to UnityPy.config.FALLBACK_UNITY_VERSION ({FALLBACK_UNITY_VERSION})",  # noqa: E501
        category=UnityVersionFallbackWarning,
        stacklevel=2,
    )

    return FALLBACK_UNITY_VERSION



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic, Optional, TypeVar, cast

from attr import define

if TYPE_CHECKING:
    from ..enums.ClassIDType import ClassIDType
    from ..files.ObjectReader import ObjectReader
    from ..files.SerializedFile import SerializedFile

T = TypeVar("T")


@define(slots=True, kw_only=True)
class PPtr(Generic[T]):
    m_FileID: int
    m_PathID: int
    assetsfile: Optional[SerializedFile] = None

    @property
    def file_id(self) -> int:
        # backwards compatibility
        return self.m_FileID

    @property
    def path_id(self) -> int:
        # backwards compatibility
        return self.m_PathID

    @property
    def type(self) -> ClassIDType:
        return self.deref().type

    # backwards compatibility - to be removed in UnityPy 2
    def read(self):
        return self.deref_parse_as_object()

    # backwards compatibility - to be removed in UnityPy 2
    def read_typetree(self):
        return self.deref_parse_as_dict()

    def deref(self, assetsfile: Optional[SerializedFile] = None) -> ObjectReader[T]:
        assetsfile = assetsfile or self.assetsfile

        if assetsfile is None:
            raise ValueError("PPtr can't deref without an assetsfile!")

        if self.m_PathID == 0:
            raise ValueError("PPtr can't deref with m_PathID == 0!")

        assetsfile_dst = None
        if self.m_FileID == 0:
            assetsfile_dst = assetsfile or self.assetsfile
        else:
            # resolve file id to external name
            external_id = self.m_FileID - 1
            if external_id >= len(assetsfile.externals):
                raise FileNotFoundError("Failed to resolve pointer - invalid m_FileID!")
            external = assetsfile.externals[external_id]

            # resolve external name to assetsfile
            container = assetsfile.parent
            if container is None:
                # TODO - use default fs
                raise FileNotFoundError(f"PPtr points to {external.path} but no container is set!")

            external_clean_path = external.path
            if external_clean_path.startswith("archive:/"):
                external_clean_path = external_clean_path[9:]
            if external_clean_path.startswith("assets/"):
                external_clean_path = external_clean_path[7:]
            external_clean_path = external_clean_path.rsplit("/")[-1].lower()

            for key, file in container.files.items():
                if key.lower() == external_clean_path:
                    assetsfile_dst = file
                    break
            else:
                env = assetsfile.environment
                cab = env.find_file(external_clean_path)
                if cab:
                    assetsfile_dst = cab
                else:
                    raise FileNotFoundError(f"Failed to resolve pointer - {external.path} not found!")

        if assetsfile_dst is None:
            raise FileNotFoundError(f"Failed to resolve pointer - {self.m_FileID} not found!")

        return cast("ObjectReader[T]", assetsfile_dst.objects[self.m_PathID])

    def deref_parse_as_object(self, assetsfile: Optional[SerializedFile] = None) -> T:
        return self.deref(assetsfile).parse_as_object()

    def deref_parse_as_dict(self, assetsfile: Optional[SerializedFile] = None) -> dict[str, Any]:
        return self.deref(assetsfile).parse_as_dict()

    def __bool__(self):
        return self.m_PathID != 0

    def __hash__(self) -> int:
        return hash((self.m_FileID, self.m_PathID))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PPtr):
            return False
        return self.m_FileID == other.m_FileID and self.m_PathID == other.m_PathID


__all__ = [
    "PPtr",
]
